# Copyright 2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.clients.blockstorage.rest_api import BlockStorageRestClient


class BlockStorageClient(BlockStorageRestClient):
    """OpenStack Block Storage v2 client"""

    def list_volumes(self, detail=None):
        """:returns: (list)"""
        r = self.volumes_get(detail=detail)
        return r.json['volumes']

    def get_volume_details(self, volume_id):
        """:returns: (dict)"""
        r = self.volumes_get(volume_id=volume_id)
        return r.json['volume']

    def create_volume(
            self, size,
            availability_zone=None,
            source_volid=None,
            display_name=None,
            display_description=None,
            snapshot_id=None,
            imageRef=None,
            volume_type=None,
            bootable=None,
            metadata=None):
        """:returns: (dict) new volumes' details"""
        r = self.volumes_post(
            size,
            availability_zone=availability_zone,
            source_volid=source_volid,
            display_name=display_name,
            display_description=display_description,
            snapshot_id=snapshot_id,
            imageRef=imageRef,
            volume_type=volume_type,
            bootable=bootable,
            metadata=metadata)
        return r.json['volume']

    def update_volume(
            self, volume_id,
            display_name=None,
            display_description=None,
            delete_on_termination=None,
            metadata=None):
        """:returns: (dict) volumes' new details"""
        args = (
            display_description, display_name, delete_on_termination, metadata)
        if args == (None, None, None, None):
            return self.get_volume_details(volume_id)
        r = self.volumes_put(
            volume_id,
            display_name=display_name,
            display_description=display_description,
            delete_on_termination=delete_on_termination,
            metadata=metadata)
        return r.json['volume']

    def delete_volume(self, volume_id):
        r = self.volumes_delete(volume_id)
        return r.headers

    def list_snapshots(self, detail=None):
        """:returns: (list)"""
        r = self.snapshots_get(detail=detail)
        return r.json['snapshots']

    def get_snapshot_details(self, snapshot_id):
        """:returns: (dict)"""
        r = self.snapshots_get(snapshot_id=snapshot_id)
        return r.json['snapshot']

    def create_snapshot(
            self, volume_id,
            force=None, display_name=None, display_description=None):
        """:returns: (dict) new snapshots' details"""
        r = self.snapshots_post(
            volume_id,
            force=force,
            display_name=display_name,
            display_description=display_description)
        return r.json['snapshot']

    def update_snapshot(
            self, snapshot_id, display_name=None, display_description=None):
        """:returns: (dict) snapshots' new details"""
        if (display_name, display_description) == (None, None):
            return self.get_snapshot_details(snapshot_id)
        r = self.snapshots_put(
            snapshot_id,
            display_name=display_name, display_description=display_description)
        return r.json['snapshot']

    def delete_snapshot(self, snapshot_id):
        r = self.snapshots_delete(snapshot_id)
        return r.headers

    def list_volume_types(self):
        r = self.types_get()
        return r.json['volume_types']

    def get_volume_type_details(self, type_id):
        r = self.types_get(type_id)
        return r.json['volume_type']



# Copyright 2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.cli.config import Config
from kamaki.clients import astakos, cyclades, ClientError
from kamaki.clients.utils import https

https.patch_with_certs('/etc/ssl/certs/ca-certificates.crt')
cnf = Config()
CLOUD = cnf.get('global', 'default_cloud')
URL = cnf.get_cloud(CLOUD, 'url')
TOKEN = cnf.get_cloud(CLOUD, 'token')
identity_client = astakos.CachedAstakosClient(URL, TOKEN)

networkURL = identity_client.get_endpoint_url(
    cyclades.CycladesNetworkClient.service_type)
network_client = cyclades.CycladesNetworkClient(networkURL, TOKEN)

ip, srv = dict(id=567262, floating_ip_address='123.45.67.89'), dict(id=123456)


def assert_status(new, expected):
    assert new is not None, 'Timeout while waiting for status %s' % expected
    assert new == expected, 'Server did not reach status %s' % expected

port = network_client.create_port(
    network_id=ip['floating_network_id'],
    device_id=srv['id'],
    fixed_ips=[dict(ip_address=ip['floating_ip_address']), ])

print 'Attaching IP', ip['floating_ip_address'], 'to server', port['device_id']
port['status'] = network_client.wait_port(port['id'], port['status'])
assert_status(port['status'], 'ACTIVE')
print '... OK'

network_client.delete_port(port['id'])

print 'Detaching IP', ip['floating_ip_address'], 'from', port['device_id']
port['status'] = network_client.wait_port(port['id'], port['status'])
assert_status(port['status'], 'DELETED')
print '... OK'



# Copyright 2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from mock import patch, call
from unittest import TestCase
from itertools import product

from kamaki.clients import blockstorage

clients_pkg = 'kamaki.clients.Client'
utils_pkg = 'kamaki.clients.utils'


class BlockStorageRestClient(TestCase):
    """Block Storage (cinder) REST API unit tests"""

    def setUp(self):
        self.url = 'http://volumes.example.com'
        self.token = 'v01um3s70k3n'
        self.client = blockstorage.BlockStorageRestClient(self.url, self.token)

    @patch('%s.get' % clients_pkg)
    @patch('%s.path4url' % utils_pkg, return_value='path_str')
    def test_volumes_get(self, path4url, get):
        for detail, volume_id, kwargs in product(
                (True, None),
                ('v0lum3-1d', None),
                ({'k': 'v'}, {'success': 1000}, {})):
            self.client.volumes_get(detail, volume_id, **kwargs)
            self.assertEqual(
                path4url.mock_calls[-1],
                call('volumes', 'detail' if detail else volume_id or ''))
            success = kwargs.pop('success', 200)
            self.assertEqual(
                get.mock_calls[-1],
                call('path_str', success=success, **kwargs))

    @patch('%s.post' % clients_pkg)
    @patch('%s.path4url' % utils_pkg, return_value='path_str')
    def test_volumes_post(self, path4url, post):
        keys = (
            'availability_zone', 'source_volid', 'display_description',
            'snapshot_id', 'display_name', 'imageRef', 'volume_type',
            'bootable', 'metadata')
        for args in product(
                ('az', None), ('volId', None), ('dd', None),
                ('sn', None), ('dn', None), ('ir', None), ('vt', None),
                (True, False, None), ({'mk': 'mv'}, None),
                ({'k1': 'v1', 'k2': 'v2'}, {'success': 1000}, {})):
            kwargs = args[-1]
            args = args[:-1]
            for err, size in ((TypeError, None), (ValueError, 'size')):
                self.assertRaises(
                    err, self.client.volumes_post, size, *args, **kwargs)
            size = 42
            self.client.volumes_post(size, *args, **kwargs)
            self.assertEqual(path4url.mock_calls[-1], call('volumes'))
            volume = dict(size=int(size))
            for k, v in zip(keys, args):
                if v is not None:
                    volume[k] = v
            success, jsondata = kwargs.pop('success', 202), dict(volume=volume)
            self.assertEqual(
                post.mock_calls[-1],
                call('path_str', json=jsondata, success=success, **kwargs))

    @patch('%s.put' % clients_pkg)
    @patch('%s.path4url' % utils_pkg, return_value='path_str')
    def test_volumes_put(self, path4url, put):
        keys = (
            'display_description', 'display_name', 'delete_on_termination',
            'metadata')
        for args in product(
                ('dd', None), ('dn', None), (True, False, None),
                ({'mk': 'mv'}, None),
                ({'k1': 'v1', 'k2': 'v2'}, {'success': 1000}, {})):
            kwargs, volid = args[-1], 'v0lum3-1d'
            args = args[:-1]
            self.client.volumes_put(volid, *args, **kwargs)
            self.assertEqual(path4url.mock_calls[-1], call('volumes', volid))
            volume = dict()
            for k, v in zip(keys, args):
                if v is not None:
                    volume[k] = v
            success, jsondata = kwargs.pop('success', 200), dict(volume=volume)
            self.assertEqual(put.mock_calls[-1], call(
                'path_str', json=jsondata, success=success, **kwargs))

    @patch('%s.delete' % clients_pkg)
    @patch('%s.path4url' % utils_pkg, return_value='path_str')
    def test_volumes_delete(self, path4url, delete):
        volid = 'v0lum3-1d'
        for kwargs in (dict(k1='v1', k2='v2'), dict(success=1000), dict()):
            self.client.volumes_delete(volid, **kwargs)
            self.assertEqual(path4url.mock_calls[-1], call('volumes', volid))
            success = kwargs.pop('success', 202)
            self.assertEqual(delete.mock_calls[-1], call(
                'path_str', success=success, **kwargs))

    @patch('%s.get' % clients_pkg)
    @patch('%s.path4url' % utils_pkg, return_value='path_str')
    def test_snapshots_get(self, path4url, get):
        for detail, snapshot_id, kwargs in product(
                (True, None),
                ('v0lum3-1d', None),
                ({'k': 'v'}, {'success': 1000}, {})):
            self.client.snapshots_get(detail, snapshot_id, **kwargs)
            self.assertEqual(
                path4url.mock_calls[-1],
                call('snapshots', 'detail' if detail else snapshot_id or ''))
            success = kwargs.pop('success', 200)
            self.assertEqual(
                get.mock_calls[-1],
                call('path_str', success=success, **kwargs))

    @patch('%s.post' % clients_pkg)
    @patch('%s.path4url' % utils_pkg, return_value='path_str')
    def test_snapshots_post(self, path4url, post):
        keys = ('force', 'display_name', 'display_description')
        for args in product(
                (True, False, None), ('dn', None), ('dd', None),
                ({'k1': 'v1', 'k2': 'v2'}, {'success': 1000}, {})):
            kwargs, volume_id = args[-1], 'v0lum3-1d'
            args = args[:-1]
            self.client.snapshots_post(volume_id, *args, **kwargs)
            self.assertEqual(path4url.mock_calls[-1], call('snapshots'))
            sn = dict(volume_id=volume_id)
            for k, v in zip(keys, args):
                if v is not None:
                    sn[k] = v
            success, json_data = kwargs.pop('success', 202), dict(snapshot=sn)
            self.assertEqual(
                post.mock_calls[-1],
                call('path_str', json=json_data, success=success, **kwargs))

    @patch('%s.put' % clients_pkg)
    @patch('%s.path4url' % utils_pkg, return_value='path_str')
    def test_snapshots_put(self, path4url, put):
        keys = ('display_description', 'display_name')
        for args in product(
                ('dd', None), ('dn', None),
                ({'k1': 'v1', 'k2': 'v2'}, {'success': 1000}, {})):
            kwargs, volid = args[-1], 'v0lum3-1d'
            args = args[:-1]
            self.client.snapshots_put(volid, *args, **kwargs)
            self.assertEqual(path4url.mock_calls[-1], call('snapshots', volid))
            sn = dict()
            for k, v in zip(keys, args):
                if v is not None:
                    sn[k] = v
            success, json_data = kwargs.pop('success', 200), dict(snapshot=sn)
            self.assertEqual(put.mock_calls[-1], call(
                'path_str', json=json_data, success=success, **kwargs))

    @patch('%s.delete' % clients_pkg)
    @patch('%s.path4url' % utils_pkg, return_value='path_str')
    def test_snapshots_delete(self, path4url, delete):
        snid = 'sn4p5h0t-1d'
        for kwargs in (dict(k1='v1', k2='v2'), dict(success=1000), dict()):
            self.client.snapshots_delete(snid, **kwargs)
            self.assertEqual(path4url.mock_calls[-1], call('snapshots', snid))
            success = kwargs.pop('success', 202)
            self.assertEqual(delete.mock_calls[-1], call(
                'path_str', success=success, **kwargs))

    @patch('%s.get' % clients_pkg)
    @patch('%s.path4url' % utils_pkg, return_value='path_str')
    def test_types_set(self, path4url, get):
        for type_id, kwargs in product(
                ('tid', None),
                ({'k1': 'v1', 'k2': 'v2'}, {'success': 1000}, {})):
            self.client.types_get(type_id, **kwargs)
            self.assertEqual(
                path4url.mock_calls[-1], call('types', type_id or ''))
            success = kwargs.pop('success', 200)
            self.assertEqual(get.mock_calls[-1], call(
                'path_str', success=success, **kwargs))


rest_pkg = 'kamaki.clients.blockstorage.BlockStorageRestClient'


class FakeResponse:
    json = 'ret'
    headers = 'hea'


class BlockStorageClient(TestCase):

    def setUp(self):
        self.url = 'http://volumes.example.com'
        self.token = 'v01um3s70k3n'
        self.client = blockstorage.BlockStorageClient(self.url, self.token)

    def tearDown(self):
        FakeResponse.json = 'ret'

    @patch('%s.volumes_get' % rest_pkg, return_value=FakeResponse())
    def test_list_volumes(self, volumes_get):
        FakeResponse.json = dict(volumes='ret list')
        self.assertEqual(self.client.list_volumes(), 'ret list')
        self.assertEqual(volumes_get.mock_calls[-1], call(detail=None))
        self.assertEqual(self.client.list_volumes(True), 'ret list')
        self.assertEqual(volumes_get.mock_calls[-1], call(detail=True))

    @patch('%s.volumes_get' % rest_pkg, return_value=FakeResponse())
    def test_get_volume_details(self, volumes_get):
        FakeResponse.json = dict(volume='ret')
        self.assertEqual(self.client.get_volume_details('vid'), 'ret')
        self.assertEqual(volumes_get.mock_calls[-1], call(volume_id='vid'))

    @patch('%s.volumes_post' % rest_pkg, return_value=FakeResponse())
    def test_create_volume(self, volumes_post):
        keys, FakeResponse.json = (
            'availability_zone', 'source_volid', 'display_name',
            'display_description', 'snapshot_id', 'imageRef', 'volume_type',
            'bootable', 'metadata'), dict(volume='ret')
        for args in product(
                ('az', None), ('volId', None), ('dn', None),
                ('dd', None), ('sn', None), ('ir', None), ('vt', None),
                (True, False, None), ({'mk': 'mv'}, None)):
            self.assertEqual(self.client.create_volume(42, *args), 'ret')
            kwargs = dict(zip(keys, args))
            self.assertEqual(volumes_post.mock_calls[-1], call(42, **kwargs))

    @patch(
        'kamaki.clients.blockstorage.BlockStorageClient.get_volume_details',
        return_value='vd')
    @patch('%s.volumes_put' % rest_pkg, return_value=FakeResponse())
    def test_update_volume(self, volumes_put, get_volume_details):
        keys, FakeResponse.json = (
            'display_name', 'display_description', 'delete_on_termination',
            'metadata'), dict(volume='ret')
        for args in product(
                ('dn', None), ('dd', None), (True, False, None),
                ({'mk': 'v'}, None)):
            if args == (None, None, None, None):
                self.assertEqual(self.client.update_volume('vid', *args), 'vd')
            else:
                self.assertEqual(
                    self.client.update_volume('vid', *args), 'ret')
                kwargs = dict(zip(keys, args))
                self.assertEqual(
                    volumes_put.mock_calls[-1], call('vid', **kwargs))

    @patch('%s.volumes_delete' % rest_pkg, return_value=FakeResponse())
    def test_delete_volume(self, volumes_delete):
        self.assertEqual(self.client.delete_volume('vid'), 'hea')
        volumes_delete.assert_called_once_with('vid')

    @patch('%s.snapshots_get' % rest_pkg, return_value=FakeResponse())
    def test_list_snapshots(self, snapshots_get):
        FakeResponse.json = dict(snapshots='ret list')
        self.assertEqual(self.client.list_snapshots(), 'ret list')
        self.assertEqual(snapshots_get.mock_calls[-1], call(detail=None))
        self.assertEqual(self.client.list_snapshots(True), 'ret list')
        self.assertEqual(snapshots_get.mock_calls[-1], call(detail=True))

    @patch('%s.snapshots_get' % rest_pkg, return_value=FakeResponse())
    def test_get_snapshot_details(self, snapshots_get):
        FakeResponse.json = dict(snapshot='ret')
        self.assertEqual(self.client.get_snapshot_details('vid'), 'ret')
        self.assertEqual(snapshots_get.mock_calls[-1], call(snapshot_id='vid'))

    @patch('%s.snapshots_post' % rest_pkg, return_value=FakeResponse())
    def test_create_snapshot(self, snapshots_post):
        keys = ('force', 'display_name', 'display_description')
        FakeResponse.json = dict(snapshot='ret')
        for args in product((True, False, None), ('dn', None), ('dd', None)):
            self.assertEqual(self.client.create_snapshot(42, *args), 'ret')
            kwargs = dict(zip(keys, args))
            self.assertEqual(snapshots_post.mock_calls[-1], call(42, **kwargs))

    @patch(
        'kamaki.clients.blockstorage.BlockStorageClient.get_snapshot_details',
        return_value='sd')
    @patch('%s.snapshots_put' % rest_pkg, return_value=FakeResponse())
    def test_update_snapshot(self, snapshots_put, get_snapshot_details):
        keys = ('display_name', 'display_description')
        FakeResponse.json = dict(snapshot='ret')
        for args in product(('dn', None), ('dd', None)):
            if args == (None, None):
                self.assertEqual(
                    self.client.update_snapshot('sid', *args), 'sd')
            else:
                self.assertEqual(
                    self.client.update_snapshot('sid', *args), 'ret')
                kwargs = dict(zip(keys, args))
                self.assertEqual(
                    snapshots_put.mock_calls[-1], call('sid', **kwargs))

    @patch('%s.snapshots_delete' % rest_pkg, return_value=FakeResponse())
    def test_delete_snapshot(self, snapshots_delete):
        self.assertEqual(self.client.delete_snapshot('sid'), 'hea')
        snapshots_delete.assert_called_once_with('sid')

    @patch('%s.types_get' % rest_pkg, return_value=FakeResponse())
    def test_list_volume_types(self, types_get):
        FakeResponse.json = dict(volume_types='ret list')
        self.assertEqual(self.client.list_volume_types(), 'ret list')
        self.assertEqual(types_get.mock_calls[-1], call())

    @patch('%s.types_get' % rest_pkg, return_value=FakeResponse())
    def test_get_volume_type_details(self, types_get):
        FakeResponse.json = dict(volume_type='ret dict')
        self.assertEqual(
            self.client.get_volume_type_details('vtid'), 'ret dict')
        self.assertEqual(types_get.mock_calls[-1], call('vtid'))


if __name__ == '__main__':
    from sys import argv
    from kamaki.clients.test import runTestCase
    not_found = True
    if not argv[1:] or argv[1] == 'BlockStorageRestClient':
        not_found = False
        runTestCase(
            BlockStorageRestClient, 'Block Storage Rest Client', argv[2:])
    if not argv[1:] or argv[1] == 'BlockStorageClient':
        not_found = False
        runTestCase(BlockStorageClient, 'Block Storage Client', argv[2:])
    if not_found:
        print('TestCase %s not found' % argv[1])



# Copyright 2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from mock import patch, call
from unittest import TestCase
from itertools import product
from json import dumps

from kamaki.clients import network


class NetworkRestClient(TestCase):

    """Set up a ComputesRest thorough test"""
    def setUp(self):
        self.url = 'http://network.example.com'
        self.token = 'n2tw0rk70k3n'
        self.client = network.NetworkRestClient(self.url, self.token)

    def tearDown(self):
        del self.client

    def _assert(self, method_call, path, set_param=None, params=(), **kwargs):
        """Assert the REST method call is called as expected"""
        x0 = - len(params)
        for i, (k, v, c) in enumerate(params):
            self.assertEqual(set_param.mock_calls[x0 + i], call(k, v, iff=c))

        self.assertEqual(method_call.mock_calls[-1], call(path, **kwargs))

    @patch('kamaki.clients.Client.get', return_value='ret val')
    def test_networks_get(self, get):
        netid = 'netid'
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            self.assertEqual(self.client.networks_get(**kwargs), 'ret val')
            self._assert(get, '/networks', **kwargs)
            self.assertEqual(
                self.client.networks_get(network_id=netid, **kwargs),
                'ret val')
            self._assert(get, '/networks/%s' % netid, **kwargs)

    @patch('kamaki.clients.Client.post', return_value='ret val')
    def test_networks_post(self, post):
        for kwargs in (
                dict(json_data=dict(k1='v1')),
                dict(json_data=dict(k2='v2'), k3='v3')):
            self.assertEqual(self.client.networks_post(**kwargs), 'ret val')
            json_data = kwargs.pop('json_data')
            self._assert(post, '/networks', json=json_data, **kwargs)

    @patch('kamaki.clients.Client.put', return_value='ret val')
    def test_networks_put(self, put):
        netid = 'netid'
        for kwargs in (
                dict(json_data=dict(k1='v1')),
                dict(json_data=dict(k2='v2'), k3='v3')):
            self.assertEqual(
                self.client.networks_put(netid, **kwargs), 'ret val')
            json_data = kwargs.pop('json_data')
            self._assert(
                put, '/networks/%s' % netid, json=json_data, **kwargs)

    @patch('kamaki.clients.Client.delete', return_value='ret val')
    def test_networks_delete(self, delete):
        netid = 'netid'
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            self.assertEqual(
                self.client.networks_delete(netid, **kwargs), 'ret val')
            self._assert(delete, '/networks/%s' % netid, **kwargs)

    @patch('kamaki.clients.Client.get', return_value='ret val')
    def test_subnets_get(self, get):
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            self.assertEqual(self.client.subnets_get(**kwargs), 'ret val')
            self._assert(get, '/subnets', **kwargs)

            subnet_id = 'subnet id'
            self.assertEqual(
                self.client.subnets_get(subnet_id=subnet_id, **kwargs),
                'ret val')
            self._assert(get, '/subnets/%s' % subnet_id, **kwargs)

    @patch('kamaki.clients.Client.post', return_value='ret val')
    def test_subnets_post(self, post):
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            json_data = dict(subnets='some data')
            self.assertEqual(self.client.subnets_post(
                json_data=json_data, **kwargs), 'ret val')
            self._assert(post, '/subnets', json=json_data, **kwargs)

    @patch('kamaki.clients.Client.put', return_value='ret val')
    def test_subnets_put(self, put):
        subnet_id = 'subid'
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            self.assertEqual(
                self.client.subnets_put(subnet_id, **kwargs), 'ret val')
            self._assert(put, '/subnets/%s' % subnet_id, **kwargs)

    @patch('kamaki.clients.Client.delete', return_value='ret val')
    def test_subnets_delete(self, delete):
        netid = 'netid'
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            self.assertEqual(
                self.client.subnets_delete(netid, **kwargs), 'ret val')
            self._assert(delete, '/subnets/%s' % netid, **kwargs)

    @patch('kamaki.clients.Client.get', return_value='ret val')
    def test_ports_get(self, get):
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            self.assertEqual(self.client.ports_get(**kwargs), 'ret val')
            self._assert(get, '/ports', **kwargs)

            port_id = 'port id'
            self.assertEqual(
                self.client.ports_get(port_id=port_id, **kwargs),
                'ret val')
            self._assert(get, '/ports/%s' % port_id, **kwargs)

    @patch('kamaki.clients.Client.post', return_value='ret val')
    def test_ports_post(self, post):
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            self.assertEqual(self.client.ports_post(**kwargs), 'ret val')
            self._assert(post, '/ports', json=None, **kwargs)

            json_data = dict(id='some id', other_param='other val')
            self.assertEqual(
                self.client.ports_post(json_data=json_data, **kwargs),
                'ret val')
            self._assert(post, '/ports', json=json_data, **kwargs)

    @patch('kamaki.clients.Client.put', return_value='ret val')
    def test_ports_put(self, put):
        port_id = 'portid'
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            self.assertEqual(
                self.client.ports_put(port_id, **kwargs), 'ret val')
            self._assert(put, '/ports/%s' % port_id, json=None, **kwargs)

            json_data = dict(id='some id', other_param='other val')
            self.assertEqual(
                self.client.ports_put(port_id, json_data=json_data, **kwargs),
                'ret val')
            self._assert(put, '/ports/%s' % port_id, json=json_data, **kwargs)

    @patch('kamaki.clients.Client.get', return_value='ret val')
    def test_floatingips_get(self, get):
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            self.assertEqual(self.client.floatingips_get(**kwargs), 'ret val')
            self._assert(get, '/floatingips', **kwargs)

            floatingip_id = 'port id'
            self.assertEqual(
                self.client.floatingips_get(
                    floatingip_id=floatingip_id, **kwargs),
                'ret val')
            self._assert(get, '/floatingips/%s' % floatingip_id, **kwargs)

    @patch('kamaki.clients.Client.post', return_value='ret val')
    def test_floatingips_post(self, post):
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            json_data = dict(id='some id', other_param='other val')
            self.assertEqual(
                self.client.floatingips_post(json_data=json_data, **kwargs),
                'ret val')
            self._assert(post, '/floatingips', json=json_data, **kwargs)

    @patch('kamaki.clients.Client.put', return_value='ret val')
    def test_floatingips_put(self, put):
        floatingip_id = 'portid'
        for kwargs in (dict(), dict(k1='v1'), dict(k2='v2', k3='v3')):
            json_data = dict(id='some id', other_param='other val')
            self.assertEqual(
                self.client.floatingips_put(
                    floatingip_id, json_data=json_data, **kwargs),
                'ret val')
            self._assert(
                put, '/floatingips/%s' % floatingip_id,
                json=json_data, **kwargs)


class FakeObject(object):

    json = None
    headers = None


class NetworkClient(TestCase):

    """Set up a ComputesRest thorough test"""
    def setUp(self):
        self.url = 'http://network.example.com'
        self.token = 'n2tw0rk70k3n'
        self.client = network.NetworkClient(self.url, self.token)

    def tearDown(self):
        FakeObject.json, FakeObject.headers = None, None
        del self.client

    @patch(
        'kamaki.clients.network.NetworkClient.networks_get',
        return_value=FakeObject())
    def test_list_networks(self, networks_get):
        FakeObject.json = dict(networks='ret val')
        self.assertEqual(self.client.list_networks(), 'ret val')
        networks_get.assert_called_once_with(success=200)

    @patch(
        'kamaki.clients.network.NetworkClient.networks_post',
        return_value=FakeObject())
    def test_create_network(self, networks_post):
        for admin_state_up, shared in product((None, True), (None, True)):
            FakeObject.json = dict(network='ret val')
            name = 'net name'
            self.assertEqual(
                self.client.create_network(
                    name, admin_state_up=admin_state_up, shared=shared),
                'ret val')
            req = dict(name=name, admin_state_up=bool(admin_state_up))
            if shared:
                req['shared'] = shared
            expargs = dict(json_data=dict(network=req), success=201)
            self.assertEqual(networks_post.mock_calls[-1], call(**expargs))

    @patch(
        'kamaki.clients.network.NetworkClient.networks_post',
        return_value=FakeObject())
    def test_create_networks(self, networks_post):
        for networks in (
                None, dict(name='name'), 'nets', [1, 2, 3], [{'k': 'v'}, ],
                [dict(admin_state_up=True, shared=True)],
                [dict(name='n1', invalid='mistake'), ],
                [dict(name='valid', shared=True), {'err': 'nop'}]):
            self.assertRaises(
                ValueError, self.client.create_networks, networks)

        FakeObject.json = dict(networks='ret val')
        for networks in (
                [
                    dict(name='net1'),
                    dict(name='net 2', admin_state_up=False, shared=True)],
                [
                    dict(name='net1', admin_state_up=True),
                    dict(name='net 2', shared=False),
                    dict(name='net-3')],
                (dict(name='n.e.t'), dict(name='net 2'))):
            self.assertEqual(self.client.create_networks(networks), 'ret val')

            networks = list(networks)
            expargs = dict(json_data=dict(networks=networks), success=201)
            self.assertEqual(networks_post.mock_calls[-1], call(**expargs))

    @patch(
        'kamaki.clients.network.NetworkClient.networks_get',
        return_value=FakeObject())
    def test_get_network_details(self, networks_get):
        netid, FakeObject.json = 'netid', dict(network='ret val')
        self.assertEqual(self.client.get_network_details(netid), 'ret val')
        networks_get.assert_called_once_with(netid, success=200)

    @patch(
        'kamaki.clients.network.NetworkClient.networks_put',
        return_value=FakeObject())
    def test_update_network(self, networks_put):
        netid, FakeObject.json = 'netid', dict(network='ret val')
        for name, admin_state_up, shared in product(
                ('net name', None), (True, None), (True, None)):
            kwargs = dict(
                name=name, admin_state_up=admin_state_up, shared=shared)
            self.assertEqual(
                self.client.update_network(netid, **kwargs), 'ret val')
            if name in (None, ):
                kwargs.pop('name')
            if admin_state_up in (None, ):
                kwargs.pop('admin_state_up')
            if shared in (None, ):
                kwargs.pop('shared')
            kwargs = dict(json_data=dict(network=kwargs), success=200)
            self.assertEqual(
                networks_put.mock_calls[-1], call(netid, **kwargs))

    @patch(
        'kamaki.clients.network.NetworkClient.networks_delete',
        return_value=FakeObject())
    def test_delete_network(self, networks_delete):
        netid, FakeObject.headers = 'netid', 'ret headers'
        self.assertEqual(self.client.delete_network(netid), 'ret headers')
        networks_delete.assert_called_once_with(netid, success=204)

    @patch(
        'kamaki.clients.network.NetworkClient.subnets_get',
        return_value=FakeObject())
    def test_list_subnets(self, subnets_get):
        FakeObject.json = dict(subnets='ret val')
        self.assertEqual(self.client.list_subnets(), 'ret val')
        subnets_get.assert_called_once_with(success=200)

    @patch(
        'kamaki.clients.network.NetworkClient.subnets_post',
        return_value=FakeObject())
    def test_create_subnet(self, subnets_post):
        for (
                name, allocation_pools, gateway_ip,
                subnet_id, ipv6, enable_dhcp) in product(
                    ('name', None), ('all pools', None), ('gip', None),
                    ('sid', None), (True, None), (True, None)):
            kwargs = dict(
                name=name, allocation_pools=allocation_pools,
                gateway_ip=gateway_ip, subnet_id=subnet_id,
                ipv6=ipv6, enable_dhcp=enable_dhcp)
            FakeObject.json, network_id, cidr = dict(subnet='rv'), 'name', 'cd'
            self.assertEqual(
                self.client.create_subnet(network_id, cidr, **kwargs), 'rv')
            req = dict(
                network_id=network_id, cidr=cidr,
                ip_version=6 if kwargs.pop('ipv6', None) else 4)
            for k, v in kwargs.items():
                if v:
                    req['id' if k == 'subnet_id' else k] = v
            expargs = dict(json_data=dict(subnet=req), success=201)
            self.assertEqual(subnets_post.mock_calls[-1], call(**expargs))

    @patch(
        'kamaki.clients.network.NetworkClient.subnets_post',
        return_value=FakeObject())
    def test_create_subnets(self, subnets_post):
        for subnets in (
                None, dict(network_id='name'), 'nets', [1, 2, 3], [{'k': 'v'}],
                [dict(ipv6=True, enable_dhcp=True)],
                [dict(network_id='n1', cidr='dr', invalid='mistake'), ],
                [dict(network_id='valid', cidr='valid'), {'err': 'nop'}]):
            self.assertRaises(
                ValueError, self.client.create_subnets, subnets)

        FakeObject.json = dict(subnets='ret val')
        for subnets in (
                [
                    dict(network_id='n1', cidr='c1'),
                    dict(network_id='n 2', cidr='c 2', name='name')],
                [
                    dict(network_id='n1', cidr='c 6', allocation_pools='a p'),
                    dict(network_id='n 2', cidr='c_4', gateway_ip='g ip'),
                    dict(network_id='n 2', cidr='c_4', subnet_id='s id'),
                    dict(network_id='n-4', cidr='c3', ipv6=True, name='w. 6'),
                    dict(network_id='n_5', cidr='c2', enable_dhcp=True)],
                (
                    dict(network_id='n.e.t', cidr='c-5'),
                    dict(network_id='net 2', cidr='c 2'))):
            self.assertEqual(self.client.create_subnets(subnets), 'ret val')

            for subnet in subnets:
                subnet['ip_version'] = 6 if subnet.pop('ipv6', None) else 4
                if 'subnet_id' in subnet:
                    subnet['id'] = subnet.pop('subnet_id')
            subnets = list(subnets)
            expargs = dict(json_data=dict(subnets=subnets), success=201)
            self.assertEqual(subnets_post.mock_calls[-1], call(**expargs))

    @patch(
        'kamaki.clients.network.NetworkClient.subnets_get',
        return_value=FakeObject())
    def test_get_subnet_details(self, subnets_get):
        subid, FakeObject.json = 'subid', 'ret val'
        self.assertEqual(self.client.get_subnet_details(subid), 'ret val')
        subnets_get.assert_called_once_with(subid, success=200)

    @patch(
        'kamaki.clients.network.NetworkClient.subnets_put',
        return_value=FakeObject())
    def test_update_subnet(self, subnets_put):
        for (
                name, allocation_pools, gateway_ip,
                ipv6, enable_dhcp) in product(
                    ('name', None), ('all pools', None), ('gip', None),
                    (True, False, None), (True, False, None)):
            kwargs = dict(
                name=name, allocation_pools=allocation_pools,
                gateway_ip=gateway_ip, ipv6=ipv6, enable_dhcp=enable_dhcp)
            FakeObject.json, subnet_id = dict(subnet='rv'), 'sid'
            self.assertEqual(
                self.client.update_subnet(subnet_id, **kwargs), 'rv')
            req = dict()
            for k, v in kwargs.items():
                if v not in (None, ):
                    if k in ('ipv6', ):
                        req['ip_version'] = 6 if v else 4
                    else:
                        req[k] = v
            expargs = dict(json=dict(subnet=req), success=200)
            self.assertEqual(
                subnets_put.mock_calls[-1], call(subnet_id, **expargs))

    @patch(
        'kamaki.clients.network.NetworkClient.subnets_delete',
        return_value=FakeObject())
    def test_delete_subnet(self, subnets_delete):
        netid, FakeObject.headers = 'netid', 'ret headers'
        self.assertEqual(self.client.delete_subnet(netid), 'ret headers')
        subnets_delete.assert_called_once_with(netid, success=204)

    @patch(
        'kamaki.clients.network.NetworkClient.ports_get',
        return_value=FakeObject())
    def test_list_ports(self, ports_get):
        FakeObject.json = dict(ports='ret val')
        self.assertEqual(self.client.list_ports(), 'ret val')
        ports_get.assert_called_once_with(success=200)

    @patch(
        'kamaki.clients.network.NetworkClient.ports_post',
        return_value=FakeObject())
    def test_create_port(self, ports_post):
        for (
                name, status, admin_state_up,
                mac_address, fixed_ips, security_groups
                ) in product(
                    ('name', None), ('status', None), (True, False, None),
                    ('maddr', None), ('some ips', None), ([1, 2, 3], None)):
            kwargs = dict(
                name=name, status=status, admin_state_up=admin_state_up,
                mac_address=mac_address, fixed_ips=fixed_ips,
                security_groups=security_groups)
            FakeObject.json, network_id = dict(port='ret val'), 'name'
            self.assertEqual(
                self.client.create_port(network_id, **kwargs), 'ret val')
            req = dict(network_id=network_id)
            for k, v in kwargs.items():
                if v not in (None, ):
                    req[k] = v
            expargs = dict(json_data=dict(port=req), success=201)
            self.assertEqual(ports_post.mock_calls[-1], call(**expargs))

    @patch(
        'kamaki.clients.network.NetworkClient.ports_post',
        return_value=FakeObject())
    def test_create_ports(self, ports_post):
        for ports in (
                None, dict(network_id='name'), 'nets', [1, 2, 3], [{'k': 'v'}],
                [dict(name=True, mac_address='mac')],
                [dict(network_id='n1', invalid='mistake'), ],
                [dict(network_id='valid', name='valid'), {'err': 'nop'}]):
            self.assertRaises(
                ValueError, self.client.create_ports, ports)

        FakeObject.json = dict(ports='ret val')
        for ports in (
                [dict(network_id='n1'), dict(network_id='n 2', name='name')],
                [
                    dict(network_id='n1', name='n 6', status='status'),
                    dict(network_id='n 2', admin_state_up=True, fixed_ips='f'),
                    dict(network_id='n 2', mac_address='mc', name='n.a.m.e.'),
                    dict(network_id='n-4', security_groups='s~G', name='w. 6'),
                    dict(network_id='n_5', admin_state_up=False, name='f a')],
                (
                    dict(network_id='n.e.t', name='c-5'),
                    dict(network_id='net 2', status='YEAH'))):
            self.assertEqual(self.client.create_ports(ports), 'ret val')
            expargs = dict(json_data=dict(ports=list(ports)), success=201)
            self.assertEqual(ports_post.mock_calls[-1], call(**expargs))

    @patch(
        'kamaki.clients.network.NetworkClient.ports_get',
        return_value=FakeObject())
    def test_get_port_details(self, ports_get):
        portid, FakeObject.json = 'portid', dict(port='ret val')
        self.assertEqual(self.client.get_port_details(portid), 'ret val')
        ports_get.assert_called_once_with(portid, success=200)

    @patch(
        'kamaki.clients.network.NetworkClient.ports_delete',
        return_value=FakeObject())
    def test_delete_port(self, ports_delete):
        portid, FakeObject.headers = 'portid', 'ret headers'
        self.assertEqual(self.client.delete_port(portid), 'ret headers')
        ports_delete.assert_called_once_with(portid, success=204)

    @patch(
        'kamaki.clients.network.NetworkClient.ports_put',
        return_value=FakeObject())
    def test_update_port(self, ports_put):
        for (
                name, status, admin_state_up, mac_address, fixed_ips,
                security_groups) in product(
                    ('name', None), ('st', None), (True, None), ('mc', None),
                    ('fps', None), ('sg', None)):
            FakeObject.json = dict(port='rv')
            port_id, network_id = 'pid', 'nid'
            kwargs = dict(
                network_id=network_id, name=name, status=status,
                admin_state_up=admin_state_up, mac_address=mac_address,
                fixed_ips=fixed_ips, security_groups=security_groups)
            self.assertEqual(
                self.client.update_port(port_id, **kwargs), 'rv')
            req = dict()
            for k, v in kwargs.items():
                if v:
                    req[k] = v
            expargs = dict(json_data=dict(port=req), success=200)
            self.assertEqual(
                ports_put.mock_calls[-1], call(port_id, **expargs))


if __name__ == '__main__':
    from sys import argv
    from kamaki.clients.test import runTestCase
    not_found = True
    if not argv[1:] or argv[1] == 'NetworkClient':
        not_found = False
        runTestCase(NetworkClient, 'Network Client', argv[2:])
    if not argv[1:] or argv[1] == 'NetworkRestClient':
        not_found = False
        runTestCase(NetworkRestClient, 'NetworkRest Client', argv[2:])
    if not_found:
        print('TestCase %s not found' % argv[1])



# Copyright 2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

import logging
import httplib
import socket
import ssl
from objpool import http

log = logging.getLogger(__name__)


class SSLUnicodeError(ssl.SSLError):
    """SSL module cannot handle unicode file names"""


class HTTPSClientAuthConnection(httplib.HTTPSConnection):
    """HTTPS connection, with full client-based SSL Authentication support"""

    ca_file, ignore_ssl = None, False

    def __init__(self, *args, **kwargs):
        """ Extent HTTPSConnection to support SSL authentication
            :param ca_file: path to CA certificates bundle (default: None)
            :param ignore_ssl: flag (default: False)
        """
        try:
            self.ca_file = str(kwargs.pop('ca_file', self.ca_file))
        except UnicodeError as ue:
            raise SSLUnicodeError(0, SSLUnicodeError.__doc__, ue)

        self.ignore_ssl = kwargs.pop('ignore_ssl', self.ignore_ssl)

        httplib.HTTPSConnection.__init__(self, *args, **kwargs)

    def connect(self):
        """Connect to a host on a given (SSL) port.
        Use ca_file to check Server Certificate.

        Redefined/copied and extended from httplib.py:1105 (Python 2.6.x).
        This is needed to pass cert_reqs=ssl.CERT_REQUIRED as parameter to
        ssl.wrap_socket(), which forces SSL to check server certificate against
        our client certificate.
        """
        source_address = getattr(self, 'source_address', None)
        socket_args = [(self.host, self.port), self.timeout] + (
            [source_address, ] if source_address else [])
        sock = socket.create_connection(*socket_args)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()

        try:
            if self.ignore_ssl:
                self.sock = ssl.wrap_socket(
                    sock, self.key_file, self.cert_file,
                    cert_reqs=ssl.CERT_NONE)
            else:
                self.sock = ssl.wrap_socket(
                    sock, self.key_file, self.cert_file,
                    ca_certs=self.ca_file, cert_reqs=ssl.CERT_REQUIRED)
        except UnicodeError as ue:
            raise SSLUnicodeError(0, SSLUnicodeError.__doc__, ue)


http.HTTPConnectionPool._scheme_to_class['https'] = HTTPSClientAuthConnection
PooledHTTPConnection = http.PooledHTTPConnection


def patch_with_certs(ca_file):
    try:
        HTTPSClientAuthConnection.ca_file = str(ca_file)
    except UnicodeError as ue:
            raise SSLUnicodeError(0, SSLUnicodeError.__doc__, ue)


def patch_ignore_ssl(insecure_connection=True):
    HTTPSClientAuthConnection.ignore_ssl = insecure_connection



# Copyright 2011-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.clients import Client, ClientError
from kamaki.clients.utils import filter_in, filter_out, path4url


class StorageClient(Client):
    """OpenStack Object Storage API 1.0 client"""

    def __init__(self, endpoint_url, token, account=None, container=None):
        super(StorageClient, self).__init__(endpoint_url, token)
        self.account = account
        self.container = container

    def _assert_account(self):
        if not self.account:
            raise ClientError("No account provided")

    def _assert_container(self):
        self._assert_account()
        if not self.container:
            raise ClientError("No container provided")

    def get_account_info(self):
        """
        :returns: (dict)
        """
        self._assert_account()
        path = path4url(self.account)
        r = self.head(path, success=(204, 401))
        if r.status_code == 401:
            raise ClientError("No authorization", status=401)
        reply = r.headers
        return reply

    def replace_account_meta(self, metapairs):
        """
        :param metapais: (dict) key:val metadata pairs
        """
        self._assert_account()
        path = path4url(self.account)
        for key, val in metapairs.items():
            self.set_header('X-Account-Meta-' + key, val)
        self.post(path, success=202)

    def del_account_meta(self, metakey):
        """
        :param metakey: (str) metadatum key
        """
        headers = self.get_account_info()
        self.headers = filter_out(
            headers,
            'X-Account-Meta-' + metakey,
            exactMatch=True)
        if len(self.headers) == len(headers):
            raise ClientError('X-Account-Meta-%s not found' % metakey, 404)
        path = path4url(self.account)
        self.post(path, success=202)

    def create_container(self, container):
        """
        :param container: (str)

        :raises ClientError: 202 Container already exists
        """
        self._assert_account()
        path = path4url(self.account, container)
        r = self.put(path, success=(201, 202))
        if r.status_code == 202:
            raise ClientError("Container already exists", r.status_code)

    def get_container_info(self, container):
        """
        :param container: (str)

        :returns: (dict)

        :raises ClientError: 404 Container does not exist
        """
        self._assert_account()
        path = path4url(self.account, container)
        r = self.head(path, success=(204, 404))
        if r.status_code == 404:
            raise ClientError("Container does not exist", r.status_code)
        reply = r.headers
        return reply

    def delete_container(self, container):
        """
        :param container: (str)

        :raises ClientError: 404 Container does not exist
        :raises ClientError: 409 Container not empty
        """
        self._assert_account()
        path = path4url(self.account, container)
        r = self.delete(path, success=(204, 404, 409))
        if r.status_code == 404:
            raise ClientError("Container does not exist", r.status_code)
        elif r.status_code == 409:
            raise ClientError("Container is not empty", r.status_code)

    def list_containers(self):
        """
        :returns: (dict)
        """
        self._assert_account()
        self.set_param('format', 'json')
        path = path4url(self.account)
        r = self.get(path, success=(200, 204))
        reply = r.json
        return reply

    def upload_object(self, obj, f, size=None):
        """ A simple (naive) implementation.

        :param obj: (str)

        :param f: an open for reading file descriptor

        :param size: (int) number of bytes to upload
        """
        self._assert_container()
        path = path4url(self.account, self.container, obj)
        data = f.read(size) if size else f.read()
        self.put(path, data=data, success=201)

    def create_object(
            self, obj,
            content_type='application/octet-stream', content_length=0):
        """
        :param obj: (str) directory-object name

        :param content_type: (str) explicitly set content_type

        :param content_length: (int) explicitly set content length

        :returns: (dict) object creation headers
        """
        self._assert_container()
        path = path4url(self.account, self.container, obj)
        self.set_header('Content-Type', content_type)
        self.set_header('Content-length', str(content_length))
        r = self.put(path, success=201)
        return r.headers

    def create_directory(self, obj):
        """
        :param obj: (str) directory-object name

        :returns: (dict) request headers
        """
        self._assert_container()
        assert obj, 'Remote directory path is missing'
        path = path4url(self.account, self.container, obj)
        self.set_header('Content-Type', 'application/directory')
        self.set_header('Content-length', '0')
        r = self.put(path, success=201)
        return r.headers

    def get_object_info(self, obj):
        """
        :param obj: (str)

        :returns: (dict)
        """
        self._assert_container()
        path = path4url(self.account, self.container, obj)
        r = self.head(path, success=200)
        reply = r.headers
        return reply

    def get_object_meta(self, obj):
        """
        :param obj: (str)

        :returns: (dict)
        """
        r = filter_in(self.get_object_info(obj), 'X-Object-Meta-')
        reply = {}
        for (key, val) in r.items():
            metakey = key.split('-')[-1]
            reply[metakey] = val
        return reply

    def del_object_meta(self, obj, metakey):
        """
        :param obj: (str)

        :param metakey: (str) the metadatum key
        """
        self._assert_container()
        self.set_header('X-Object-Meta-' + metakey, '')
        path = path4url(self.account, self.container, obj)
        self.post(path, success=202)

    def replace_object_meta(self, metapairs):
        """
        :param metapairs: (dict) key:val metadata
        """
        self._assert_container()
        path = path4url(self.account, self.container)
        for key, val in metapairs.items():
            self.set_header('X-Object-Meta-' + key, val)
        self.post(path, success=202)

    def copy_object(
            self, src_container, src_object, dst_container,
            dst_object=False):
        """Copy an objects from src_contaier:src_object to
            dst_container[:dst_object]

        :param src_container: (str)

        :param src_object: (str)

        :param dst_container: (str)

        :param dst_object: (str)
        """
        self._assert_account()
        dst_object = dst_object or src_object
        dst_path = path4url(self.account, dst_container, dst_object)
        self.set_header('X-Copy-From', path4url(src_container, src_object))
        self.set_header('Content-Length', 0)
        self.put(dst_path, success=201)

    def move_object(
            self, src_container, src_object, dst_container,
            dst_object=False):
        """Move an objects from src_contaier:src_object to
            dst_container[:dst_object]

        :param src_container: (str)

        :param src_object: (str)

        :param dst_container: (str)

        :param dst_object: (str)
        """
        self._assert_account()
        dst_object = dst_object or src_object
        dst_path = path4url(self.account, dst_container, dst_object)
        self.set_header('X-Move-From', path4url(src_container, src_object))
        self.set_header('Content-Length', 0)
        self.put(dst_path, success=201)

    def delete_object(self, obj):
        """
        :param obj: (str)

        :raises ClientError: 404 Object not found
        """
        self._assert_container()
        path = path4url(self.account, self.container, obj)
        r = self.delete(path, success=(204, 404))
        if r.status_code == 404:
            raise ClientError("Object %s not found" % obj, r.status_code)

    def list_objects(
            self,
            limit=None,
            marker=None,
            prefix=None,
            format=None,
            delimiter=None,
            path=None):
        """
        :param limit: (integer) The amount of results requested

        :param marker: (string) Return containers with name lexicographically
            after marker

        :param prefix: (string) Return objects starting with prefix

        :param format: (string) reply format can be json or xml (default:json)

        :param delimiter: (string) Return objects up to the delimiter

        :param path: (string) assume prefix = path and delimiter = /
            (overwrites prefix and delimiter)

        :returns: (dict)

        :raises ClientError: 404 Invalid account
        """
        self._assert_container()
        restpath = path4url(self.account, self.container)

        self.set_param('format', format or 'json')

        self.set_param('limit', limit, iff=limit)
        self.set_param('marker', marker, iff=marker)
        if path:
            self.set_param('path', path)
        else:
            self.set_param('prefix', prefix, iff=prefix)
            self.set_param('delimiter', delimiter, iff=delimiter)

        r = self.get(restpath, success=(200, 204, 304, 404), )
        if r.status_code == 404:
            raise ClientError(
                "Invalid account (%s) for that container" % self.account,
                r.status_code)
        elif r.status_code == 304:
            return []
        return r.json

    def list_objects_in_path(self, path_prefix):
        """
        :param path_prefix: (str)

        :raises ClientError: 404 Invalid account

        :returns: (dict)
        """
        self._assert_container()
        path = path4url(self.account, self.container)
        self.set_param('format', 'json')
        self.set_param('path', path_prefix)
        r = self.get(path, success=(200, 204, 404))
        if r.status_code == 404:
            raise ClientError(
                "Invalid account (%s) for that container" % self.account,
                r.status_code)
        reply = r.json
        return reply



# Copyright 2011-2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.command

from traceback import format_exc, format_stack
from logging import getLogger
from astakosclient import AstakosClientException

from kamaki.clients import ClientError
from kamaki.cli.errors import CLIError, CLISyntaxError
from kamaki.cli.utils import format_size

log = getLogger(__name__)

CLOUDNAME = ['Note: Set a cloud and use its name instead of "default"']


class Generic(object):

    @classmethod
    def all(this, func):
        def _raise(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except Exception as e:
                log.debug('Error stack:\n%s' % ''.join(format_stack()))
                log.debug(format_exc(e))
                if isinstance(e, CLIError):
                    raise e
                elif isinstance(e, ClientError):
                    raise CLIError(
                        '(%s) %s' % (getattr(e, 'status', 'no status'), e),
                        details=getattr(e, 'details', []),
                        importance=1 if (
                            e.status < 200) else 2 if (
                            e.status < 300) else 3 if (
                            e.status < 400) else 4)
                raise CLIError(
                    '%s' % e, details=['%s, -d for debug info' % type(e)])
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def _connection(this, func):
        def _raise(self, *args, **kwargs):
            try:
                func(self, *args, **kwargs)
            except ClientError as ce:
                ce_msg = ('%s' % ce).lower()
                if ce.status == 401:
                    raise CLIError('Authorization failed', details=[
                        'To check if token is valid',
                        '  kamaki user authenticate',
                        'To set token:',
                        '  kamaki config set cloud.default.token <token>',
                        'To get current token:',
                        '  kamaki config get cloud.default.token',
                        '%s %s' % (getattr(ce, 'status', ''), ce)] + CLOUDNAME)
                elif ce.status in range(-12, 200) + [302, 401, 500]:
                    raise CLIError(
                        '%s %s' % (getattr(ce, 'status', ''), ce),
                        importance=3)
                elif ce.status == 404 and 'kamakihttpresponse' in ce_msg:
                    client = getattr(self, 'client', None)
                    if not client:
                        raise
                    url = getattr(client, 'endpoint_url', '<empty>')
                    raise CLIError('Invalid service URL %s' % url, details=[
                        'Check if authentication URL is correct',
                        'To check current URL',
                        '  kamaki config get cloud.default.url',
                        'To set new authentication URL',
                        '  kamaki config set cloud.default.url',
                        '%s %s' % (getattr(ce, 'status', ''), ce)] + CLOUDNAME)
                raise
        _raise.__name__ = func.__name__
        return _raise


class Astakos(object):

    _token_details = [
        'To check a token:',
        '  kamaki config get cloud.CLOUD.token',
        'To see all configured clouds:',
        '  kamaki config get cloud'
        'To set/get the default cloud:',
        '  kamaki config get default_cloud',
        '  kamaki config set default_cloud CLOUD'
        'If (re)set a token:',
        '  #  (permanent)',
        '  $ kamaki config set cloud.CLOUD.token <token>']

    @classmethod
    def astakosclient(this, func):
        def _raise(self, *args, **kwargs):
            try:
                r = func(self, *args, **kwargs)
            except AstakosClientException as ace:
                raise CLIError(
                    'Error in AstakosClient', details=['%s' % ace, ])
            return r
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def project_id(this, func):
        def _raise(self, *args, **kwargs):
            project_id = kwargs.get('project_id', None)
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if project_id and ce.status in (400, 404):
                    raise CLIError(
                        'No project with ID %s' % project_id,
                        importance=2, details=[
                            'To see all projects', '  kamaki project list',
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                elif project_id and ce.status in (403, ):
                    raise CLIError(
                        'No access to project %s' % project_id,
                        importance=3, details=[
                            'To see all projects', '  kamaki project list',
                            'To see  memberships',
                            '  kamaki membership list',
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def membership_id(this, func):
        def _raise(self, *args, **kwargs):
            membership_id = kwargs.get('membership_id', None)
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if membership_id and ce.status in (400, 404):
                    raise CLIError(
                        'No membership with ID %s' % membership_id,
                        importance=2, details=[
                            'To list all memberships',
                            '  kamaki membership list',
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                elif membership_id and ce.status in (403, ):
                    raise CLIError(
                        'No access to membership %s' % membership_id,
                        importance=3, details=[
                            'To see all memberships',
                            '  kamaki membership list',
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
        _raise.__name__ = func.__name__
        return _raise


class History(object):
    @classmethod
    def init(this, func):
        def _raise(self, *args, **kwargs):
            r = func(self, *args, **kwargs)
            if not hasattr(self, 'history'):
                raise CLIError('Failed to load history', importance=2)
            return r
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def _get_cmd_ids(this, func):
        def _raise(self, cmd_ids, *args, **kwargs):
            if not cmd_ids:
                raise CLISyntaxError(
                    'Usage: <id1|id1-id2> [id3|id3-id4] ...',
                    details=self.__doc__.split('\n'))
            return func(self, cmd_ids, *args, **kwargs)
        _raise.__name__ = func.__name__
        return _raise


class Cyclades(object):
    about_flavor_id = [
        'To get a list of flavors', '  kamaki flavor list',
        'More details on a flavor', '  kamaki flavor info FLAVOR_ID', ]

    about_network_id = [
        'To get a list of networks', '  kamaki network list',
        'More details on a network', '  kamaki network info NETWORK_ID', ]

    about_ips = [
        'To list available IPs', '  kamaki ip list',
        'To reserve a new IP', '  kamaki ip create', ]

    @classmethod
    def connection(this, func):
        return Generic._connection(func)

    @classmethod
    def date(this, func):
        def _raise(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status in (304, ):
                    log.debug('%s %s' % (ce.status, ce))
                    self.error('No servers have been modified since')
                else:
                    raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def cluster_size(this, func):
        def _raise(self, *args, **kwargs):
            size = kwargs.get('size', None)
            try:
                size = int(size)
                assert size > 0, 'Cluster size must be a positive integer'
                return func(self, *args, **kwargs)
            except ValueError as ve:
                msg = 'Invalid cluster size value %s' % size
                raise CLIError(msg, importance=1, details=[
                    'Cluster size must be a positive integer', '%s' % ve])
            except AssertionError as ae:
                raise CLIError(
                    'Invalid cluster size %s' % size,
                    importance=1, details=['%s' % ae])
            except ClientError:
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def network_id(this, func):
        def _raise(self, *args, **kwargs):
            network_id = kwargs.get('network_id', None)
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if network_id and ce.status in (404, 400):
                    msg = ''
                    if ce.status in (400, ):
                        try:
                            network_id = int(network_id)
                        except ValueError:
                            msg = 'Network ID should be a positive integer'
                            log.debug(msg)
                    raise CLIError(
                        'No network with id %s found' % network_id,
                        importance=2,
                        details=[msg, ] + this.about_network_id + [
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def network_in_use(this, func):
        def _raise(self, *args, **kwargs):
            network_id = kwargs.get('network_id', None)
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status in (409, ):
                    raise CLIError(
                        'Network with id %s is in use' % network_id,
                        importance=3, details=[
                            'To list all network ports', '  kamaki port list',
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def network_permissions(this, func):
        def _raise(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status in (403, ):
                    network_id = kwargs.get('network_id', '')
                    raise CLIError(
                        'Insufficient permissions for this action',
                        importance=2, details=[
                            'To get information on network',
                            '  kamaki network info %s' % network_id,
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def subnet_id(this, func):
        def _raise(self, *args, **kwargs):
            subnet_id = kwargs.get('subnet_id', None)
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if subnet_id and ce.status in (404, 400):
                    details = []
                    if ce.status in (400, ):
                        try:
                            subnet_id = int(subnet_id)
                        except ValueError:
                            details = ['Subnet ID should be positive integer']
                            log.debug(details[-1])
                    raise CLIError(
                        'No subnet with id %s found' % subnet_id,
                        importance=2, details=details + [
                            'To list subnets', '  kamaki subnet list',
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def subnet_permissions(this, func):
        def _raise(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status in (401, ):
                    subnet_id = kwargs.get('subnet_id', '')
                    raise CLIError(
                        'Insufficient permissions for this action',
                        importance=2, details=[
                            'Make sure this subnet belongs to current user',
                            'To see information on subnet',
                            '  kamaki subnet info %s' % subnet_id,
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def port_id(this, func):
        def _raise(self, *args, **kwargs):
            port_id = kwargs.get('port_id', None)
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if port_id and ce.status in (404, 400):
                    details = []
                    if ce.status in (400, ):
                        try:
                            port_id = int(port_id)
                        except ValueError:
                            details = ['Port ID should be positive integer']
                            log.debug(details[-1])
                    raise CLIError(
                        'No port with id %s found' % port_id,
                        importance=2, details=details + [
                            'To list ports', '  kamaki port list',
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def ip_id(this, func):
        def _raise(self, *args, **kwargs):
            ip_id = kwargs.get('ip_id', None)
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status in (404, 400):
                    details = []
                    if ce.status in (400, ):
                        try:
                            ip_id = int(ip_id)
                        except ValueError:
                            details = ['IP ID should be positive integer']
                            log.debug(details[-1])
                    raise CLIError(
                        'No floating IP with ID %s found' % ip_id,
                        importance=2, details=details + this.about_ips + [
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def flavor_id(this, func):
        def _raise(self, *args, **kwargs):
            flavor_id = kwargs.get('flavor_id', None)
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status in (404, 400):
                    details = this.about_flavor_id
                    if ce.status in (400, ):
                        try:
                            flavor_id = int(flavor_id)
                        except ValueError:
                            details.insert(
                                0, 'Flavor ID should be a positive integer')
                    raise CLIError(
                        'No flavor with ID %s' % flavor_id,
                        importance=2, details=details + [
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def server_id(this, func):
        def _raise(self, *args, **kwargs):
            details = ['To get a list of all servers', '  kamaki server list']
            server_id = kwargs.get('server_id', None)
            try:
                server_id = int(server_id)
                assert server_id > 0, 'error: %s is not positive' % server_id
            except (ValueError, AssertionError) as err:
                raise CLIError(
                    'Invalid server id %s' % server_id,
                    importance=2, details=[
                        'Server id must be a positive integer'] + details + [
                            err, ])
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status in (404, ):
                    raise CLIError(
                        'No servers with ID %s' % server_id,
                        importance=2, details=details + [
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def metadata(this, func):
        def _raise(self, *args, **kwargs):
            key = kwargs.get('key', None)
            try:
                func(self, *args, **kwargs)
            except ClientError as ce:
                if key and ce.status == 404 and (
                        'metadata' in ('%s' % ce).lower()):
                    raise CLIError(
                        'No virtual server metadata with key %s' % key,
                        details=['%s %s' % (getattr(ce, 'status', ''), ce), ])
                raise
        _raise.__name__ = func.__name__
        return _raise


class Image(object):
    about_image_id = [
        'To list all images', '  kamaki image list',
        'To get image details', '  kamaki image info IMAGE_ID', ]

    @classmethod
    def connection(this, func):
        return Generic._connection(func)

    @classmethod
    def id(this, func):
        def _raise(self, *args, **kwargs):
            image_id = kwargs.get('image_id', None)
            try:
                func(self, *args, **kwargs)
            except ClientError as ce:
                if image_id and ce.status in (404, 400):
                    raise CLIError(
                        'No image with id %s found' % image_id,
                        importance=2, details=this.about_image_id + [
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def permissions(this, func):
        def _raise(self, *args, **kwargs):
            try:
                func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status in (403, 405):
                    raise CLIError(
                        'Insufficient permissions for this action',
                        importance=2, details=[
                            'To see the owner of an image',
                            '  kamaki image info IMAGE_ID',
                            'To see image file permissions',
                            '  kamaki file info IMAGE_LOCATION --sharing',
                            '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise


class Pithos(object):
    container_howto = [
        'To list containers',
        '  kamaki container list',
        'Hint: Use a / to refer to a container (default: /pithos) e.g.,',
        'To list contents of container "images"',
        '  kamaki file list /images',
        'To get information on file "my.img" in container "images"',
        '  kamaki file info /images/my.img', ]

    @classmethod
    def connection(this, func):
        return Generic._connection(func)

    @classmethod
    def account(this, func):
        def _raise(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status == 403:
                    raise CLIError(
                        'Insufficient credentials for this operation',
                        details=['%s %s' % (getattr(ce, 'status', ''), ce), ])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def quota(this, func):
        def _raise(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status == 413:
                    raise CLIError('User quota exceeded', details=[
                        'To get total quotas',
                        '  kamaki quota list --resource=pithos',
                        'To get container limit',
                        '  kamaki container info CONTAINER --size-limit',
                        'Set a higher container limit:',
                        '  kamaki container modify CONTAINER '
                        '--size-limit=NEW_LIMIT',
                        '%s' % ce])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def container(this, func):
        def _raise(self, *args, **kwargs):
            dst_cont = kwargs.get('dst_cont', None)
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status in (404, ):
                        cont = ('%s or %s' % (self.container, dst_cont)) if (
                            dst_cont) else self.container
                        raise CLIError(
                            'Container "%s" does not exist' % cont,
                            importance=2, details=this.container_howto + [
                                '%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def local_path_download(this, func):
        def _raise(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except IOError as ioe:
                raise CLIError(
                    'Failed to access a local file', importance=2, details=[
                        'To check if the file exists',
                        '  kamaki file info PATH',
                        'All directories in a remote path must exist, or the '
                        'download will fail',
                        'To create a remote directory',
                        '  kamaki file mkdir REMOTE_DIRECTORY_PATH',
                        u'%s' % ioe])
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def local_path(this, func):
        def _raise(self, *args, **kwargs):
            local_path = kwargs.get('local_path', None)
            try:
                return func(self, *args, **kwargs)
            except IOError as ioe:
                raise CLIError(
                    'Failed to access file %s' % local_path,
                    details=[u'%s' % ioe, ], importance=2)
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def object_path(this, func):
        def _raise(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if ce.status in (404, ):
                    _cnt = self.container
                    _cnt = '[/%s]' % _cnt if _cnt == 'pithos' else '/%s' % _cnt
                    raise CLIError(
                        'No object "%s" in container "%s"' % (
                            self.path, self.container),
                        importance=2, details=[
                            'To list contents in container',
                            '  kamaki file list %s' % _cnt,
                            '%s %s' % (getattr(ce, 'status', ''), ce), ])
                raise
        _raise.__name__ = func.__name__
        return _raise

    @classmethod
    def object_size(this, func):
        def _raise(self, *args, **kwargs):
            size = kwargs.get('size', None)
            start = kwargs.get('start', 0)
            end = kwargs.get('end', 0)
            if size:
                try:
                    size = int(size)
                except ValueError as ve:
                    raise CLIError(
                        'Invalid file size %s ' % size,
                        details=['size must be a positive integer', '%s' % ve],
                        importance=1)
            else:
                try:
                    start = int(start)
                except ValueError as e:
                    raise CLIError(
                        'Invalid start value %s in range' % start,
                        details=['size must be a positive integer', '%s' % e],
                        importance=1)
                try:
                    end = int(end)
                except ValueError as e:
                    raise CLIError(
                        'Invalid end value %s in range' % end,
                        details=['size must be a positive integer', '%s' % e],
                        importance=1)
                if start > end:
                    raise CLIError(
                        'Invalid range %s-%s' % (start, end),
                        details=['size must be a positive integer'],
                        importance=1)
                size = end - start
            try:
                return func(self, *args, **kwargs)
            except ClientError as ce:
                if size and ce.status in (416, 400):
                    raise CLIError(
                        'Remote object %s:%s <= %s %s' % (
                            self.container, self.path, format_size(size),
                            ('(%sB)' % size) if size >= 1024 else ''),
                        details=['%s %s' % (getattr(ce, 'status', ''), ce)])
                raise
        _raise.__name__ = func.__name__
        return _raise



# Copyright 2012-2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.cli.cmdtree import CommandTree
from kamaki.cli.argument import ValueArgument
from kamaki.cli.history import History
from kamaki.cli import command
from kamaki.cli.cmds import CommandInit, errors


history_cmds = CommandTree('history', 'Kamaki command history')
namespaces = [history_cmds]


class _HistoryInit(CommandInit):
    @errors.Generic.all
    @errors.History.init
    def _run(self):
        self.history = History(self.config.get('global', 'history_file'))
        self.history.limit = self.config.get('global', 'history_limit')

    def main(self):
        self._run()


@command(history_cmds)
class history_show(_HistoryInit):
    """Show history
        Featutes:
        - slice notation (cmd numbers --> N or :N or N: or N1:N2)
        - text matching (--match)
    """

    arguments = dict(
        match=ValueArgument('Show lines matching this', '--match'),
    )

    @errors.Generic.all
    def _run(self, cmd_slice):
        c = self.history.counter
        lines = ['%s.  %s' % (i + c, l) for i, l in enumerate(
            self.history[:])][cmd_slice]
        if not isinstance(cmd_slice, slice):
            lines = [lines, ]
        if self['match']:
            lines = [l for l in lines if self.history._match(l, self['match'])]
        self.print_items([l[:-1] for l in lines])

    def main(self, cmd_numbers=''):
        super(self.__class__, self)._run()
        sl_args = [
            int(x) if x else None for x in cmd_numbers.split(':')] if (
                cmd_numbers) else [None, None]
        slice_cmds = slice(*sl_args) if len(sl_args) > 1 else sl_args[0]
        self._run(slice_cmds)


@command(history_cmds)
class history_clean(_HistoryInit):
    """Clean up history (permanent)"""

    @errors.Generic.all
    def _run(self):
        self.history.empty()

    def main(self):
        super(self.__class__, self)._run()
        self._run()



# Copyright 2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.cli.config import Config

cnf = Config()
CLOUD = cnf.get('global', 'default_cloud')
URL = cnf.get_cloud(CLOUD, 'url')
TOKEN = cnf.get_cloud(CLOUD, 'token')


from kamaki.cli import logger

logger.add_stream_logger('kamaki.clients.recv', fmt='< %(message)s')
logger.add_stream_logger('kamaki.clients.send', fmt='> %(message)s')


from kamaki.clients import astakos, pithos, cyclades, image
from kamaki.clients.utils import https

https.patch_with_certs('/etc/ssl/certs/ca-certificates.crt')

identity_client = astakos.AstakosClient(URL, TOKEN)

pithosURL = identity_client.get_endpoint_url(pithos.PithosClient.service_type)
storage_client = pithos.PithosClient(pithosURL, TOKEN)
storage_client.account = identity_client.user_info['id']
storage_client.container = 'pithos'

imageURL = identity_client.get_endpoint_url(image.ImageClient.service_type)
image_client = image.ImageClient(imageURL, TOKEN)

computeURL = identity_client.get_endpoint_url(
    cyclades.CycladesComputeClient.service_type)
compute_client = cyclades.CycladesComputeClient(computeURL, TOKEN)

networkURL = identity_client.get_endpoint_url(
    cyclades.CycladesNetworkClient.service_type)
network_client = cyclades.CycladesNetworkClient(networkURL, TOKEN)

volumeURL = identity_client.get_endpoint_url(
    cyclades.CycladesBlockStorageClient.service_type)
volume_client = cyclades.CycladesBlockStorageClient(volumeURL, TOKEN)



# Copyright 2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.clients import Client, utils


class BlockStorageRestClient(Client):
    """Block Storage (cinder) REST API"""
    service_type = 'volume'

    def volumes_get(self, detail=None, volume_id=None, success=200, **kwargs):
        """GET endpoint_url/volumes[/<detail> | /<volume_id>]

        :param detail: (boolean)
        :param volume_id: (str)
        """
        path = utils.path4url(
            'volumes', 'detail' if detail else volume_id or '')
        return self.get(path, success=success, **kwargs)

    def volumes_post(
            self, size,
            availability_zone=None,
            source_volid=None,
            display_description=None,
            snapshot_id=None,
            display_name=None,
            imageRef=None,
            volume_type=None,
            bootable=None,
            metadata=None,
            success=202,
            **kwargs):
        """POST endpoint_url/volumes

        :param size: (int) Volume size in GBs

        :param availability_zone: (str)
        :param source_volid: (str) ID for existing volume to create from
        :param display_description: (str)
        :param snapshot_id: (str) ID of existing snapshot to create from
        :param display_name: (str)
        :param imageRef: (str) ID of image to create from
        :param volume_type: (str)
        :param bootable: (boolean)
        :param metadata: (dict) {metakey: metavalue, ...}
        """
        path = utils.path4url('volumes')
        volume = dict(size=int(size))
        if availability_zone is not None:
            volume['availability_zone'] = availability_zone
        if source_volid is not None:
            volume['source_volid'] = source_volid
        if display_description is not None:
            volume['display_description'] = display_description
        if snapshot_id is not None:
            volume['snapshot_id'] = snapshot_id
        if display_name is not None:
            volume['display_name'] = display_name
        if imageRef is not None:
            volume['imageRef'] = imageRef
        if volume_type is not None:
            volume['volume_type'] = volume_type
        if bootable is not None:
            volume['bootable'] = bootable
        if metadata is not None:
            volume['metadata'] = metadata
        return self.post(
            path, json=dict(volume=volume), success=success, **kwargs)

    def volumes_put(
            self, volume_id,
            display_description=None,
            display_name=None,
            delete_on_termination=None,
            metadata=None,
            success=200,
            **kwargs):
        """PUT endpoint_url/volumes/volume_id

        :param volume_id: (str)

        :param display_description: (str)
        :param display_name: (str)
        :param metadata: (dict) {metakey: metavalue, ...}
        """
        path = utils.path4url('volumes', volume_id)
        volume = dict()
        if display_description is not None:
            volume['display_description'] = display_description
        if display_name is not None:
            volume['display_name'] = display_name
        if delete_on_termination is not None:
            volume['delete_on_termination'] = bool(delete_on_termination)
        if metadata is not None:
            volume['metadata'] = metadata
        return self.put(
            path, json=dict(volume=volume), success=success, **kwargs)

    def volumes_delete(self, volume_id, success=202, **kwargs):
        """DEL ETE endpoint_url/volumes/volume_id
        :param volume_id: (str)
        """
        return self.delete(
            utils.path4url('volumes', volume_id), success=success, **kwargs)

    def snapshots_get(
            self, detail=None, snapshot_id=None, success=200, **kwargs):
        """GET endpoint_url/snapshots[/<detail> | /<snapshot_id>]

        :param detail: (boolean)

        :param snapshot_id: (str)
        """
        path = utils.path4url(
            'snapshots', 'detail' if detail else snapshot_id or '')
        return self.get(path, success=success, **kwargs)

    def snapshots_post(
            self, volume_id,
            force=None,
            display_name=None,
            display_description=None,
            success=202,
            **kwargs):
        """POST endpoint_url/snapshots

        :param volume_id: (str)

        :param force: (boolean)
        :param display_name: (str)
        :param display_description: (str)
        """
        path = utils.path4url('snapshots')
        snapshot = dict(volume_id=volume_id)
        if force is not None:
            snapshot['force'] = bool(force)
        if display_name is not None:
            snapshot['display_name'] = display_name
        if display_description is not None:
            snapshot['display_description'] = display_description
        return self.post(
            path, json=dict(snapshot=snapshot), success=success, **kwargs)

    def snapshots_put(
            self, snapshot_id,
            display_description=None,
            display_name=None,
            success=200,
            **kwargs):
        """PUT endpoint_url/snapshots/spapshot_id

        :param snapshot_id: (str)

        :param display_description: (str)
        :param display_name: (str)
        """
        path = utils.path4url('snapshots', snapshot_id)
        snapshot = dict()
        if display_description is not None:
            snapshot['display_description'] = display_description
        if display_name is not None:
            snapshot['display_name'] = display_name
        return self.put(
            path, json=dict(snapshot=snapshot), success=success, **kwargs)

    def snapshots_delete(self, snapshot_id, success=202, **kwargs):
        """DEL ETE endpoint_url/snapshots/snapshot_id
        :param snapshot_id: (str)
        """
        path = utils.path4url('snapshots', snapshot_id)
        return self.delete(path, success=success, **kwargs)

    def types_get(self, type_id=None, success=200, **kwargs):
        """GET endpoint_url/types[/<type_id>]"""
        path = utils.path4url('types', type_id or '')
        return self.get(path, success=success, **kwargs)



# Copyright 2011-2014 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from io import StringIO
from pydoc import pager

from kamaki.cli import command
from kamaki.cli.cmdtree import CommandTree
from kamaki.cli.errors import CLIInvalidArgument, raiseCLIError
from kamaki.clients.cyclades import (
    CycladesNetworkClient, ClientError, CycladesComputeClient)
from kamaki.cli.argument import (
    FlagArgument, ValueArgument, RepeatableArgument, IntArgument,
    StatusArgument)
from kamaki.cli.cmds import (
    CommandInit, OptionalOutput, NameFilter, IDFilter, errors, client_log)
from kamaki.cli.cmds import Wait


network_cmds = CommandTree('network', 'Network API network commands')
port_cmds = CommandTree('port', 'Network API port commands')
subnet_cmds = CommandTree('subnet', 'Network API subnet commands')
ip_cmds = CommandTree('ip', 'Network API floatingip commands')
namespaces = [network_cmds, port_cmds, subnet_cmds, ip_cmds]

port_states = ('BUILD', 'ACTIVE', 'DOWN', 'ERROR')


class _PortWait(Wait):

    def wait(self, port_id, current_status, timeout=60):
        super(_PortWait, self).wait(
            'Port', port_id, self.client.wait_port, current_status,
            timeout=timeout)


class _NetworkInit(CommandInit):
    @errors.Generic.all
    @client_log
    def _run(self):
        self.client = self.get_client(CycladesNetworkClient, 'network')

    def _filter_by_user_id(self, nets):
        return [net for net in nets if net['user_id'] == self['user_id']] if (
            self['user_id']) else nets

    def _get_compute_client(self):
        compute = getattr(self, '_compute_client', None)
        if not compute:
            compute = self.get_client(CycladesComputeClient, 'cyclades')
            self._compute_client = compute
        return compute

    @errors.Cyclades.network_id
    def _network_exists(self, network_id):
        self.client.get_network_details(network_id)

    @errors.Cyclades.server_id
    def _server_exists(self, server_id):
        compute_client = self._get_compute_client()
        compute_client.get_server_details(server_id)

    def _ip_exists(self, ip, network_id, error):
        for ip_item in self.client.list_floatingips():
            if ip_item['floating_ip_address'] == ip:
                if network_id and ip_item['floating_network_id'] != network_id:
                    raiseCLIError(error, details=[
                        'Floating IP %s does not belong to network %s ,' % (
                            ip, network_id),
                        'To get information on IP %s' % ip,
                        '  kamaki ip info %s' % ip_item['id']])
                return
        raiseCLIError(error, details=[
            'Floating IP %s not found' % ip] + errors.Cyclades.about_ips)

    def main(self):
        self._run()


@command(network_cmds)
class network_list(_NetworkInit, OptionalOutput, NameFilter, IDFilter):
    """List networks
    Use filtering arguments (e.g., --name-like) to manage long lists
    """

    arguments = dict(
        detail=FlagArgument('show detailed output', ('-l', '--details')),
        more=FlagArgument(
            'output results in pages (-n to set items per page, default 10)',
            '--more'),
        user_id=ValueArgument(
            'show only networks belonging to user with this id', '--user-id')
    )

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self):
        nets = self.client.list_networks(detail=True)
        nets = self._filter_by_user_id(nets)
        nets = self._filter_by_name(nets)
        nets = self._filter_by_id(nets)
        if not self['detail']:
            nets = [dict(
                id=n['id'],
                name=n['name'],
                public='( %s )' % ('public' if (
                    n.get('public', None)) else 'private')) for n in nets]
            kwargs = dict(title=('id', 'name', 'public'))
        else:
            kwargs = dict()
        if self['more']:
            kwargs['out'] = StringIO()
            kwargs['title'] = ()
        self.print_(nets, **kwargs)
        if self['more']:
            pager(kwargs['out'].getvalue())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(network_cmds)
class network_info(_NetworkInit, OptionalOutput):
    """Get details about a network"""

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.network_id
    def _run(self, network_id):
        net = self.client.get_network_details(network_id)
        self.print_(net, self.print_dict)

    def main(self, network_id):
        super(self.__class__, self)._run()
        self._run(network_id=network_id)


class NetworkTypeArgument(ValueArgument):

    types = ('MAC_FILTERED', 'CUSTOM', 'IP_LESS_ROUTED', 'PHYSICAL_VLAN')

    @property
    def value(self):
        return getattr(self, '_value', self.types[0])

    @value.setter
    def value(self, new_value):
        if new_value and new_value.upper() in self.types:
            self._value = new_value.upper()
        elif new_value:
            raise CLIInvalidArgument(
                'Invalid network type %s' % new_value, details=[
                    'Valid types: %s' % ', '.join(self.types), ])


@command(network_cmds)
class network_create(_NetworkInit, OptionalOutput):
    """Create a new network (default type: MAC_FILTERED)"""

    arguments = dict(
        name=ValueArgument('Network name', '--name'),
        shared=FlagArgument(
            'Make network shared (special privileges required)', '--shared'),
        project_id=ValueArgument('Assign network to project', '--project-id'),
        network_type=NetworkTypeArgument(
            'Valid network types: %s' % (', '.join(NetworkTypeArgument.types)),
            '--type')
    )

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self):
        try:
            net = self.client.create_network(
                self['network_type'],
                name=self['name'],
                shared=self['shared'],
                project_id=self['project_id'])
        except ClientError as ce:
            if self['project_id'] and ce.status in (400, 403, 404):
                self._project_id_exists(project_id=self['project_id'])
            raise
        self.print_(net, self.print_dict)

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(network_cmds)
class network_reassign(_NetworkInit, OptionalOutput):
    """Assign a network to a different project"""

    arguments = dict(
        project_id=ValueArgument('Assign network to project', '--project-id'),
    )
    required = ('project_id', )

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.network_permissions
    @errors.Cyclades.network_id
    def _run(self, network_id):
        try:
            self.client.reassign_network(network_id, self['project_id'])
        except ClientError as ce:
            if ce.status in (400, 403, 404):
                self._project_id_exists(project_id=self['project_id'])
            raise

    def main(self, network_id):
        super(self.__class__, self)._run()
        self._run(network_id=network_id)


@command(network_cmds)
class network_delete(_NetworkInit):
    """Delete a network"""

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.network_permissions
    @errors.Cyclades.network_in_use
    @errors.Cyclades.network_id
    def _run(self, network_id):
        self.client.delete_network(network_id)

    def main(self, network_id):
        super(self.__class__, self)._run()
        self._run(network_id=network_id)


@command(network_cmds)
class network_modify(_NetworkInit, OptionalOutput):
    """Modify network attributes"""

    arguments = dict(new_name=ValueArgument('Rename the network', '--name'))
    required = ['new_name', ]

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.network_permissions
    @errors.Cyclades.network_id
    def _run(self, network_id):
        r = self.client.update_network(network_id, name=self['new_name'])
        self.print_(r, self.print_dict)

    def main(self, network_id):
        super(self.__class__, self)._run()
        self._run(network_id=network_id)


@command(subnet_cmds)
class subnet_list(_NetworkInit, OptionalOutput, NameFilter, IDFilter):
    """List subnets
    Use filtering arguments (e.g., --name-like) to manage long server lists
    """

    arguments = dict(
        detail=FlagArgument('show detailed output', ('-l', '--details')),
        more=FlagArgument('output results in pages', '--more')
    )

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self):
        nets = self.client.list_subnets()
        nets = self._filter_by_name(nets)
        nets = self._filter_by_id(nets)
        if not self['detail']:
            nets = [dict(
                id=n['id'],
                name=n['name'],
                net='( of network %s )' % n['network_id']) for n in nets]
            kwargs = dict(title=('id', 'name', 'net'))
        else:
            kwargs = dict()
        if self['more']:
            kwargs['out'] = StringIO()
            kwargs['title'] = ()
        self.print_(nets, **kwargs)
        if self['more']:
            pager('%s' % kwargs['out'].getvalue())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(subnet_cmds)
class subnet_info(_NetworkInit, OptionalOutput):
    """Get details about a subnet"""

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.subnet_id
    def _run(self, subnet_id):
        net = self.client.get_subnet_details(subnet_id)
        self.print_(net, self.print_dict)

    def main(self, subnet_id):
        super(self.__class__, self)._run()
        self._run(subnet_id=subnet_id)


class AllocationPoolArgument(RepeatableArgument):

    @property
    def value(self):
        return super(AllocationPoolArgument, self).value or []

    @value.setter
    def value(self, new_pools):
        if not new_pools:
            return
        new_list = []
        for pool in new_pools:
            start, comma, end = pool.partition(',')
            if not (start and comma and end):
                raise CLIInvalidArgument(
                    'Invalid allocation pool argument %s' % pool, details=[
                    'Allocation values must be of the form:',
                    '  <start address>,<end address>'])
            new_list.append(dict(start=start, end=end))
        self._value = new_list


@command(subnet_cmds)
class subnet_create(_NetworkInit, OptionalOutput):
    """Create a new subnet"""

    arguments = dict(
        name=ValueArgument('Subnet name', '--name'),
        allocation_pools=AllocationPoolArgument(
            'start_address,end_address of allocation pool (can be repeated)'
            ' e.g., --alloc-pool=123.45.67.1,123.45.67.8',
            '--alloc-pool'),
        gateway=ValueArgument('Gateway IP', '--gateway'),
        subnet_id=ValueArgument('The id for the subnet', '--id'),
        ipv6=FlagArgument('If set, IP version is set to 6, else 4', '--ipv6'),
        enable_dhcp=FlagArgument('Enable dhcp (default: off)', '--with-dhcp'),
        network_id=ValueArgument('Set the network ID', '--network-id'),
        cidr=ValueArgument('Set the CIDR', '--cidr')
    )
    required = ('network_id', 'cidr')

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self):
        try:
            net = self.client.create_subnet(
                self['network_id'], self['cidr'],
                self['name'], self['allocation_pools'], self['gateway'],
                self['subnet_id'], self['ipv6'], self['enable_dhcp'])
        except ClientError as ce:
            if ce.status in (404, 400):
                self._network_exists(network_id=self['network_id'])
            raise
        self.print_(net, self.print_dict)

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(subnet_cmds)
class subnet_modify(_NetworkInit, OptionalOutput):
    """Modify the attributes of a subnet"""

    arguments = dict(
        new_name=ValueArgument('New name of the subnet', '--name')
    )
    required = ['new_name']

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.subnet_permissions
    @errors.Cyclades.subnet_id
    def _run(self, subnet_id):
        r = self.client.update_subnet(subnet_id, name=self['new_name'])
        self.print_(r, self.print_dict)

    def main(self, subnet_id):
        super(self.__class__, self)._run()
        self._run(subnet_id=subnet_id)


@command(port_cmds)
class port_list(_NetworkInit, OptionalOutput, NameFilter, IDFilter):
    """List all ports"""

    arguments = dict(
        detail=FlagArgument('show detailed output', ('-l', '--details')),
        more=FlagArgument('output results in pages', '--more'),
        user_id=ValueArgument(
            'show only networks belonging to user with this id', '--user-id')
    )

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self):
        ports = self.client.list_ports()
        ports = self._filter_by_user_id(ports)
        ports = self._filter_by_name(ports)
        ports = self._filter_by_id(ports)
        if not self['detail']:
            ports = [dict(id=p['id'], name=p['name']) for p in ports]
        kwargs = dict()
        if self['more']:
            kwargs['out'] = StringIO()
            kwargs['title'] = ()
        self.print_(ports, **kwargs)
        if self['more']:
            pager(kwargs['out'].getvalue())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(port_cmds)
class port_info(_NetworkInit, OptionalOutput):
    """Get details about a port"""

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.port_id
    def _run(self, port_id):
        port = self.client.get_port_details(port_id)
        self.print_(port, self.print_dict)

    def main(self, port_id):
        super(self.__class__, self)._run()
        self._run(port_id=port_id)


@command(port_cmds)
class port_delete(_NetworkInit, _PortWait):
    """Delete a port (== disconnect server from network)"""

    arguments = dict(
        wait=FlagArgument('Wait port to be deleted', ('-w', '--wait'))
    )

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.port_id
    def _run(self, port_id):
        if self['wait']:
            status = self.client.get_port_details(port_id)['status']
        self.client.delete_port(port_id)
        if self['wait']:
            try:
                self.wait(port_id, status)
            except ClientError as ce:
                if ce.status not in (404, ):
                    raise
                self.error('Port %s is deleted' % port_id)

    def main(self, port_id):
        super(self.__class__, self)._run()
        self._run(port_id=port_id)


@command(port_cmds)
class port_modify(_NetworkInit, OptionalOutput):
    """Modify the attributes of a port"""

    arguments = dict(new_name=ValueArgument('New name of the port', '--name'))
    required = ['new_name', ]

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.port_id
    def _run(self, port_id):
        r = self.client.get_port_details(port_id)
        r = self.client.update_port(
            port_id, r['network_id'], name=self['new_name'])
        self.print_(r, self.print_dict)

    def main(self, port_id):
        super(self.__class__, self)._run()
        self._run(port_id=port_id)


class _port_create(_NetworkInit, OptionalOutput, _PortWait):

    @errors.Cyclades.subnet_id
    def _subnet_exists(self, subnet_id):
        self.client.get_subnet_details(subnet_id)

    def connect(self, network_id, device_id):
        subnet_id, ip = self['subnet_id'], self['ip_address']
        fixed_ips = [dict(ip_address=ip)] if (ip) else None
        if fixed_ips and subnet_id:
            fixed_ips[0]['subnet_id'] = subnet_id
        try:
            r = self.client.create_port(
                network_id, device_id,
                name=self['name'],
                security_groups=self['security_group_id'],
                fixed_ips=fixed_ips)
        except ClientError as ce:
            if ce.status in (400, 404):
                self._network_exists(network_id=network_id)
                self._server_exists(server_id=device_id)
                if subnet_id:
                    self._subnet_exists(subnet_id=subnet_id)
                if self['ip_address']:
                    self._ip_exists(ip=ip, network_id=network_id, error=ce)
            raise
        if self['wait']:
            self.wait(r['id'], r['status'])
            r = self.client.get_port_details(r['id'])
        self.print_([r])


@command(port_cmds)
class port_create(_port_create):
    """Create a new port (== connect server to network)"""

    arguments = dict(
        name=ValueArgument('A human readable name', '--name'),
        security_group_id=RepeatableArgument(
            'Add a security group id (can be repeated)',
            ('-g', '--security-group')),
        subnet_id=ValueArgument(
            'Subnet id for fixed ips (used with --ip-address)',
            '--subnet-id'),
        ip_address=ValueArgument('IP address for subnet id', '--ip-address'),
        network_id=ValueArgument('Set the network ID', '--network-id'),
        device_id=ValueArgument(
            'The device is either a virtual server or a virtual router',
            '--device-id'),
        wait=FlagArgument('Wait port to be established', ('-w', '--wait')),
    )
    required = ('network_id', 'device_id')

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self):
        self.connect(self['network_id'], self['device_id'])

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(port_cmds)
class port_wait(_NetworkInit, _PortWait):
    """Wait for port to finish (default: BUILD)"""

    arguments = dict(
        port_status=StatusArgument(
            'Wait while in this status (%s, default: %s)' % (
                ', '.join(port_states), port_states[0]),
            '--status',
            valid_states=port_states),
        timeout=IntArgument(
            'Wait limit in seconds (default: 60)', '--timeout', default=60)
    )

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self, port_id, port_status):
        port = self.client.get_port_details(port_id)
        if port['status'].lower() == port_status.lower():
            self.wait(port_id, port_status, timeout=self['timeout'])
        else:
            self.error(
                'Port %s: Cannot wait for status %s, '
                'status is already %s' % (
                    port_id, port_status, port['status']))

    def main(self, port_id):
        super(self.__class__, self)._run()
        port_status = self['port_status'] or port_states[0]
        self._run(port_id=port_id, port_status=port_status)


@command(ip_cmds)
class ip_list(_NetworkInit, OptionalOutput):
    """List reserved floating IPs"""

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self):
        self.print_(self.client.list_floatingips())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(ip_cmds)
class ip_info(_NetworkInit, OptionalOutput):
    """Get details on a floating IP"""

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.ip_id
    def _run(self, ip_id):
        self.print_(self.client.get_floatingip_details(ip_id), self.print_dict)

    def main(self, ip_id):
        super(self.__class__, self)._run()
        self._run(ip_id=ip_id)


@command(ip_cmds)
class ip_create(_NetworkInit, OptionalOutput):
    """Reserve an IP on a network"""

    arguments = dict(
        network_id=ValueArgument(
            'The network to preserve the IP on', '--network-id'),
        ip_address=ValueArgument('Allocate an IP address', '--address'),
        project_id=ValueArgument('Assign the IP to project', '--project-id'),
    )

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self):
        try:
            self.print_(
                self.client.create_floatingip(
                    self['network_id'],
                    floating_ip_address=self['ip_address'],
                    project_id=self['project_id']),
                self.print_dict)
        except ClientError as ce:
            if ce.status in (400, 404):
                network_id, ip = self['network_id'], self['ip_address']
                self._network_exists(network_id=network_id)
                if ip:
                    self._ip_exists(ip, network_id, ce)
            if self['project_id'] and ce.status in (400, 403, 404):
                self._project_id_exists(project_id=self['project_id'])
            raise

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(ip_cmds)
class ip_reassign(_NetworkInit):
    """Assign a floating IP to a different project"""

    arguments = dict(
        project_id=ValueArgument('Assign the IP to project', '--project-id'),
    )
    required = ('project_id', )

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self, ip):
        try:
            self.client.reassign_floating_ip(ip, self['project_id'])
        except ClientError as ce:
            if ce.status in (400, 404):
                self._ip_exists(ip=ip, network_id=None, error=ce)
            raise

    def main(self, IP):
        super(self.__class__, self)._run()
        self._run(ip=IP)


@command(ip_cmds)
class ip_delete(_NetworkInit):
    """Unreserve an IP (also delete the port, if attached)"""

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.ip_id
    def _run(self, ip_id):
        self.client.delete_floatingip(ip_id)

    def main(self, ip_id):
        super(self.__class__, self)._run()
        self._run(ip_id=ip_id)


@command(ip_cmds)
class ip_attach(_port_create):
    """Attach an IP on a virtual server"""

    arguments = dict(
        name=ValueArgument('A human readable name for the port', '--name'),
        security_group_id=RepeatableArgument(
            'Add a security group id (can be repeated)',
            ('-g', '--security-group')),
        subnet_id=ValueArgument('Subnet id', '--subnet-id'),
        wait=FlagArgument('Wait IP to be attached', ('-w', '--wait')),
        server_id=ValueArgument('Server to attach to this IP', '--server-id')
    )
    required = ('server_id', )

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self, ip_or_ip_id):
        netid = None
        for ip in self.client.list_floatingips():
            if ip_or_ip_id in (ip['floating_ip_address'], ip['id']):
                netid = ip['floating_network_id']
                iparg = ValueArgument(parsed_name='--ip')
                iparg.value = ip['floating_ip_address']
                self.arguments['ip_address'] = iparg
                break
        if netid:
            server_id = self['server_id']
            self.error('Creating a port to attach IP %s to server %s' % (
                ip_or_ip_id, server_id))
            try:
                self.connect(netid, server_id)
            except ClientError as ce:
                self.error('Failed to connect network %s with server %s' % (
                    netid, server_id))
                if ce.status in (400, 404):
                    self._server_exists(server_id=server_id)
                    self._network_exists(network_id=netid)
                raise
        else:
            raiseCLIError(
                '%s does not match any reserved IPs or IP ids' % ip_or_ip_id,
                details=errors.Cyclades.about_ips)

    def main(self, ip_or_ip_id):
        super(self.__class__, self)._run()
        self._run(ip_or_ip_id=ip_or_ip_id)


@command(ip_cmds)
class ip_detach(_NetworkInit, _PortWait, OptionalOutput):
    """Detach an IP from a virtual server"""

    arguments = dict(
        wait=FlagArgument('Wait network to disconnect', ('-w', '--wait')),
    )

    @errors.Generic.all
    @errors.Cyclades.connection
    def _run(self, ip_or_ip_id):
        for ip in self.client.list_floatingips():
            if ip_or_ip_id in (ip['floating_ip_address'], ip['id']):
                if not ip['port_id']:
                    raiseCLIError('IP %s is not attached' % ip_or_ip_id)
                self.error('Deleting port %s:' % ip['port_id'])
                self.client.delete_port(ip['port_id'])
                if self['wait']:
                    port_status = self.client.get_port_details(ip['port_id'])[
                        'status']
                    try:
                        self.wait(ip['port_id'], port_status)
                    except ClientError as ce:
                        if ce.status not in (404, ):
                            raise
                        self.error('Port %s is deleted' % ip['port_id'])
                return
        raiseCLIError('IP or IP id %s not found' % ip_or_ip_id)

    def main(self, ip_or_ip_id):
        super(self.__class__, self)._run()
        self._run(ip_or_ip_id)


@command(network_cmds)
class network_connect(_port_create):
    """Connect a network with a device (server or router)"""

    arguments = dict(
        name=ValueArgument('A human readable name for the port', '--name'),
        security_group_id=RepeatableArgument(
            'Add a security group id (can be repeated)',
            ('-g', '--security-group')),
        subnet_id=ValueArgument(
            'Subnet id for fixed ips (used with --ip-address)',
            '--subnet-id'),
        ip_address=ValueArgument(
            'IP address for subnet id (used with --subnet-id', '--ip-address'),
        wait=FlagArgument('Wait network to connect', ('-w', '--wait')),
        device_id=RepeatableArgument(
            'Connect this device to the network (can be repeated)',
            '--device-id')
    )
    required = ('device_id', )

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.network_id
    def _run(self, network_id, server_id):
        self.error('Creating a port to connect network %s with device %s' % (
            network_id, server_id))
        try:
            self.connect(network_id, server_id)
        except ClientError as ce:
            if ce.status in (400, 404):
                self._server_exists(server_id=server_id)
            raise

    def main(self, network_id):
        super(self.__class__, self)._run()
        for sid in self['device_id']:
            self._run(network_id=network_id, server_id=sid)


@command(network_cmds)
class network_disconnect(_NetworkInit, _PortWait, OptionalOutput):
    """Disconnect a network from a device"""

    arguments = dict(
        wait=FlagArgument('Wait network to disconnect', ('-w', '--wait')),
        device_id=RepeatableArgument(
            'Disconnect device from the network (can be repeated)',
            '--device-id')
    )
    required = ('device_id', )

    @errors.Cyclades.server_id
    def _get_vm(self, server_id):
        return self._get_compute_client().get_server_details(server_id)

    @errors.Generic.all
    @errors.Cyclades.connection
    @errors.Cyclades.network_id
    def _run(self, network_id, server_id):
        vm = self._get_vm(server_id=server_id)
        ports = [port for port in vm['attachments'] if (
            port['network_id'] in (network_id, ))]
        if not ports:
            raiseCLIError('Device %s has no network %s attached' % (
                server_id, network_id), importance=2, details=[
                    'To get device networking',
                    '  kamaki server info %s --nics' % server_id])
        for port in ports:
            if self['wait']:
                port['status'] = self.client.get_port_details(port['id'])[
                    'status']
            self.client.delete_port(port['id'])
            self.error('Deleting port %s (net-id: %s, device-id: %s):' % (
                port['id'], network_id, server_id))
            if self['wait']:
                try:
                    self.wait(port['id'], port['status'])
                except ClientError as ce:
                    if ce.status not in (404, ):
                        raise
                    self.error('Port %s is deleted' % port['id'])

    def main(self, network_id):
        super(self.__class__, self)._run()
        for sid in self['device_id']:
            self._run(network_id=network_id, server_id=sid)
